import click
import pkgutil

from inspect import getmembers
from click.core import BaseCommand
from manage.utils import import_string


def add_click_commands(module, cli, command_dict, namespaced):
    """Loads all click commands"""
    module_commands = [
        item for item in getmembers(module)
        if isinstance(item[1], BaseCommand)
    ]
    options = command_dict.get('config', {})
    namespace = command_dict.get('namespace')
    for name, function in module_commands:
        f_options = options.get(name, {})
        command_name = f_options.get('name', name)
        if namespace:
            command_name = '{}_{}'.format(namespace, command_name)
        elif namespaced:
            module_namespace = module.__name__.split('.')[-1]
            command_name = '{}_{}'.format(module_namespace, command_name)
        function.short_help = f_options.get('help_text', function.short_help)
        cli.add_command(function, name=command_name)


def make_command_from_string(code, cmd_context, options, help_text=None):
    def _command(**kwargs):
        exec (code, cmd_context, kwargs)

    if help_text:
        _command.__doc__ = help_text
    _command = click.command()(_command)
    for name, option in options.items():
        _command = click.option(name, **option)(_command)
    return _command


def make_command_from_function(function, options, help_text=None):

    if help_text:
        function.__doc__ = help_text

    function = click.command()(function)
    for name, option in options.items():
        function = click.option(name, **option)(function)
    return function


def get_context(context):
    return {item: import_string(item) for item in context}


def load_commands(cli, manage_dict):
    """Loads the commands defined in manage file"""
    namespaced = manage_dict.get('namespaced')

    # get click commands
    commands = manage_dict.get('click_commands', [])
    for command_dict in commands:
        root_module = import_string(command_dict['module'])
        if getattr(root_module, '__path__', None):
            # This is a package
            iter_modules = pkgutil.iter_modules(
                root_module.__path__, prefix=root_module.__name__ + '.'
            )
            submodules_names = [item[1] for item in iter_modules]
            submodules = [import_string(name) for name in submodules_names]
            for module in submodules:
                add_click_commands(module, cli, command_dict, namespaced)
        else:
            # a single file module
            add_click_commands(root_module, cli, command_dict, namespaced)

    # get inline commands
    commands = manage_dict.get('inline_commands', [])
    for command_dict in commands:
        name = command_dict['name']
        help_text = command_dict.get('help_text')
        options = command_dict.get('options', {})
        context = command_dict.get('context', [])
        code = command_dict['code']
        cli.add_command(
            make_command_from_string(
                code=code,
                cmd_context=get_context(context),
                options=options,
                help_text=help_text
            ),
            name=name
        )

    # get function commands
    commands = manage_dict.get('function_commands', [])
    for command_dict in commands:
        name = command_dict['name']
        help_text = command_dict.get('help_text')
        options = command_dict.get('options', {})
        function = import_string(command_dict['function'])
        cli.add_command(
            make_command_from_function(
                function=function,
                options=options,
                help_text=help_text
            ),
            name=name
        )


def load_command_sources(manager, manage_dict):
    manage_sources = manage_dict.get('command_sources', [])
    for source_data in manage_sources:
        if isinstance(source_data, dict):
            source = import_string(source_data['name'])
            manager.add_source(
                source(
                    *source_data.get('args', []),
                    **source_data.get('kwargs', {})
                )
                if callable(source)
                else source
            )
        else:
            source = import_string(source_data)
            manager.add_source(source() if callable(source) else source)



# coding: utf-8
from manage.utils import import_string


def get_name(obj, default):
    default = default.split('.')[0]
    return getattr(obj, '__name__', default)


def import_objects(manage_dict):
    auto_import = {}
    auto_scripts = []
    import_dict = manage_dict.get('shell', {}).get('auto_import', {})
    for name, spec in import_dict.get('objects', {}).items():
        _obj = import_string(name)
        if spec:
            if 'init' in spec:
                method_name = spec['init'].keys()[0]
                args = spec['init'].get(method_name, {}).get('args', [])
                kwargs = spec['init'].get(method_name, {}).get('kwargs', {})
                getattr(_obj, method_name)(*args, **kwargs)
            auto_import[spec.get('as', get_name(_obj, name))] = _obj
            if 'init_script' in spec:
                auto_scripts.append(spec['init_script'])
        else:
            auto_import[get_name(_obj, name)] = _obj
    for script in auto_scripts:
        exec(script, auto_import)
    return auto_import


def exec_init(manage_dict, context):
    for name, spec in manage_dict['shell'].get('init', {}).items():
        _obj = context.get(name, import_string(name))
        args = spec.get('args', []) if spec else []
        kwargs = spec.get('kwargs', {}) if spec else {}
        _obj(*args, **kwargs)


def exec_init_script(manage_dict, context):
    if 'init_script' in manage_dict['shell']:
        exec(manage_dict['shell']['init_script'], context)



# coding: utf-8
import sys
import importlib

PY2 = sys.version_info[0] == 2
WIN = sys.platform.startswith('win')


if PY2:
    exec('def reraise(tp, value, tb=None):\n raise tp, value, tb')
else:
    def reraise(tp, value, tb=None):
        if value.__traceback__ is not tb:
            raise value.with_traceback(tb)
        raise value


def import_string(import_name, silent=False):
    """Imports an object based on a string.  This is useful if you want to
    use import paths as endpoints or something similar.  An import path can
    be specified either in dotted notation (``xml.sax.saxutils.escape``)
    or with a colon as object delimiter (``xml.sax.saxutils:escape``).
    If `silent` is True the return value will be `None` if the import fails.
    :param import_name: the dotted name for the object to import.
    :param silent: if set to `True` import errors are ignored and
                   `None` is returned instead.
    :return: imported object
    """
    # force the import name to automatically convert to strings
    # __import__ is not able to handle unicode strings in the fromlist
    # if the module is a package
    import_name = str(import_name).replace(':', '.')
    try:
        try:
            __import__(import_name)
        except ImportError:
            if '.' not in import_name:
                raise
        else:
            return sys.modules[import_name]

        module_name, obj_name = import_name.rsplit('.', 1)
        try:
            module = __import__(module_name, None, None, [obj_name])
        except ImportError:
            # support importing modules not yet set up by the parent module
            # (or package for that matter)
            module = import_string(module_name)

        try:
            return getattr(module, obj_name)
        except AttributeError as e:
            raise ImportError(e)

    except ImportError as e:
        if not silent:
            reraise(
                ImportStringError,
                ImportStringError(import_name, e),
                sys.exc_info()[2])


class ImportStringError(ImportError):

    """Provides information about a failed :func:`import_string` attempt."""

    #: String in dotted notation that failed to be imported.
    import_name = None
    #: Wrapped exception.
    exception = None

    def __init__(self, import_name, exception):
        self.import_name = import_name
        self.exception = exception

        msg = (
            'import_string() failed for %r. Possible reasons are:\n\n'
            '- missing __init__.py in a package;\n'
            '- package or module path not included in sys.path;\n'
            '- duplicated package or module name taking precedence in '
            'sys.path;\n'
            '- missing module, class, function or variable;\n\n'
            'Debugged import:\n\n%s\n\n'
            'Original exception:\n\n%s: %s')

        name = ''
        tracked = []
        for part in import_name.replace(':', '.').split('.'):
            name += (name and '.') + part
            imported = import_string(name, silent=True)
            if imported:
                tracked.append((name, getattr(imported, '__file__', None)))
            else:
                track = ['- %r found in %r.' % (n, i) for n, i in tracked]
                track.append('- %r not found.' % name)
                msg = msg % (import_name, '\n'.join(track),
                             exception.__class__.__name__, str(exception))
                break

        ImportError.__init__(self, msg)

    def __repr__(self):
        return '<%s(%r, %r)>' % (self.__class__.__name__, self.import_name,
                                 self.exception)


class DynamicObjectImporter(object):
    def __init__(self, module_name):
        self.module_name = module_name

    def __getattr__(self, item):
        return importlib.import_module(
            '{0}.{1}'.format(self.module_name, item))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import sys
import copy
import code
import json
import yaml
import click
import readline
import rlcompleter
from manage.template import default_manage_dict
from manage.auto_import import import_objects, exec_init, exec_init_script
from manage.commands_collector import load_commands, load_command_sources

MANAGE_FILE = 'manage.yml'
HIDDEN_MANAGE_FILE = '.{0}'.format(MANAGE_FILE)
MANAGE_DICT = {}


def load_manage_dict(filename=None):
    manage_filename = None
    if not MANAGE_DICT:
        if filename:
            manage_filename = filename
        elif os.path.exists(MANAGE_FILE):
            manage_filename = MANAGE_FILE
        elif os.path.exists(HIDDEN_MANAGE_FILE):
            manage_filename = HIDDEN_MANAGE_FILE
        else:
            MANAGE_DICT.update(copy.deepcopy(default_manage_dict))
            MANAGE_DICT['shell']['banner']['message'] = (
                "WARNING: This is not a managed project\n"
                "\tPlease `exit()` and \n"
                "\trun `$ manage init`\n"
                "\tand edit `manage.yml` file with desired options"
            )
            MANAGE_DICT['shell']['auto_import']['display'] = False
        if manage_filename:
            with open(manage_filename) as manage_file:
                MANAGE_DICT.update(yaml.load(manage_file))
    return MANAGE_DICT


class Config(object):

    def __init__(self):
        self.filename = None
        self._manage_dict = None

    @property
    def manage_dict(self):
        if not self._manage_dict:
            self._manage_dict = load_manage_dict(self.filename)
        return self._manage_dict


@click.group(no_args_is_help=False)
def cli():
    """ Core commands wrapper """


@cli.command()
@click.option('--banner')
@click.option('--hidden/--no-hidden', default=False)
@click.option('--backup/--no-backup', default=True)
def init(banner, hidden, backup):
    """Initialize a manage shell in current directory
        $ manage init --banner="My awesome app shell"
        initializing manage...
        creating manage.yml
    """
    manage_file = HIDDEN_MANAGE_FILE if hidden else MANAGE_FILE
    if os.path.exists(manage_file):
        if not click.confirm('Rewrite {0}?'.format(manage_file)):
            return

        if backup:
            bck = '.bck_{0}'.format(manage_file)
            with open(manage_file, 'r') as source, open(bck, 'w') as bck_file:
                bck_file.write(source.read())

    with open(manage_file, 'w') as output:
        data = default_manage_dict
        if banner:
            data['shell']['banner']['message'] = banner
        output.write(yaml.dump(data, default_flow_style=False))


@cli.command()
def debug():
    """Shows the parsed manage file"""
    print(json.dumps(MANAGE_DICT, indent=2))


def create_shell(ipython, ptpython, manage_dict=None, extra_vars=None):
    """Creates the shell"""
    manage_dict = manage_dict or MANAGE_DICT
    _vars = globals()
    _vars.update(locals())
    auto_imported = import_objects(manage_dict)
    if extra_vars:
        auto_imported.update(extra_vars)
    _vars.update(auto_imported)
    msgs = []
    if manage_dict['shell']['banner']['enabled']:
        msgs.append(
            manage_dict['shell']['banner']['message'].format(**manage_dict)
        )
    if auto_imported and manage_dict['shell']['auto_import']['display']:
        auto_imported_names = [
            key for key in auto_imported.keys()
            if key not in ['__builtins__', 'builtins']
        ]
        msgs.append('\tAuto imported: {0}\n'.format(auto_imported_names))

    banner_msg = u'\n'.join(msgs)

    if manage_dict['shell']['readline_enabled']:
        readline.set_completer(rlcompleter.Completer(_vars).complete)
        readline.parse_and_bind('tab: complete')

    exec_init(manage_dict, _vars)
    exec_init_script(manage_dict, _vars)

    if ptpython:
        try:
            from ptpython.repl import embed
            embed({}, _vars)
        except ImportError:
            click.echo("ptpython is not installed!")
        return

    try:
        if ipython is True:
            from IPython import start_ipython
            from traitlets.config import Config
            c = Config()
            c.TerminalInteractiveShell.banner2 = banner_msg
            start_ipython(argv=[], user_ns=_vars, config=c)
        else:
            raise ImportError
    except ImportError:
        shell = code.InteractiveConsole(_vars)
        shell.interact(banner=banner_msg)


@cli.command()
@click.option('--ipython/--no-ipython', default=True)
@click.option('--ptpython', default=False, is_flag=True)
def shell(ipython, ptpython):
    """Runs a Python shell with context"""
    return create_shell(ipython, ptpython, MANAGE_DICT)


def load_manage_dict_from_sys_args():
    single_option = [item for item in sys.argv if '--managefile=' in item]
    if single_option:
        filename = single_option[0].split('=')[-1]
        sys.argv.remove(single_option[0])
    elif '--managefile' in sys.argv:
        filename = sys.argv[sys.argv.index('--managefile') + 1]
        sys.argv.remove('--managefile')
        sys.argv.remove(filename)
    else:
        filename = None
    load_manage_dict(filename)


def init_cli(cli_obj, reset=False):
    if reset:
        global MANAGE_DICT
        MANAGE_DICT = {}
    sys.path.insert(0, '.')
    load_manage_dict_from_sys_args()
    cli.help = MANAGE_DICT.get(
        'help_text', '{project_name} Interactive shell!'
    ).format(**MANAGE_DICT)
    load_commands(cli, MANAGE_DICT)
    manager = click.CommandCollection(help=cli.help, no_args_is_help=False)
    manager.add_source(cli)
    load_command_sources(manager, MANAGE_DICT)
    for item in MANAGE_DICT.get('disabled', []):
        cli.commands.pop(item, None)
    return manager


def main():
    manager = init_cli(cli)
    return manager()


if __name__ == '__main__':
    main()
